from .tables import commands as tables_commands
from covid_cloud.cli.utils import assert_config, get_config
from covid_cloud.client import *
from covid_cloud.cli.auth.utils import get_oauth_token


@click.group()
@click.pass_context
def search(ctx):
    assert_config(ctx, 'search-url')


@search.command()
@click.pass_context
@click.argument("q")
@click.option('-d', '--download', is_flag=True)
@click.option('-j', '--use-json', '--json', is_flag=True)
@click.option('-r', '--raw', is_flag=True)
def query(ctx, q, download, use_json, raw):
    click.echo(search_client.query(get_config(ctx,'search-url'), q, download, use_json, raw, get_config(ctx,'oauth_token')))


search.add_command(tables_commands.tables)



import click
import urllib3
import threading
import os
from typing import Optional
from covid_cloud.constants import *
from requests import Response
from requests.exceptions import HTTPError
from covid_cloud.client.utils import get_drs_client
from covid_cloud.client.auth import login
import gzip
import re
import sys
import io
import pandas as pd

#since our downloads are multi-threaded, we use a lock to avoid race conditions
output_lock = threading.Lock()

def get_host(url):
    return re.search(r'(?<=https://)([^/])+(?=/.*)',url).group(0)

def handle_file_response(download_file,data):
    #decode if fasta
    if re.search(r'\.fa',download_file):
        data = data.decode('utf-8')
        
    return data


def file_to_dataframe(download_file,data):
    #turn into dataframe for FASTA/FASTQ files, otherwise just return raw data
    if re.search(r'\.fa',download_file):
        data = data.split('\n',maxsplit=1)

        meta = data[0]
        sequence = data[1].replace('\n','') #remove newlines
        
        return pd.DataFrame({
            "meta": [meta],
            "sequence": [sequence]
            })
    
    return data


def is_drs_url(url):
    return url[:6] == 'drs://'


def get_object_info_url_from_drs(url):
    drs_host = re.search(r'(?<=drs://)([^/])+(?=/.*)',url)
    if not drs_host:
        click.secho(f'Could not get drs object id from url {url}\n',fg='red')
        raise Exception(f'Could not get drs object id from url {url}')
    object_url = f'https://{drs_host.group(0)}/ga4gh/drs/v1/'
    return object_url, drs_host


def download_file(url, output_dir, oauth_token: Optional[dict] = None, 
                email: Optional[str] = None, 
                personal_access_token: Optional[str] = None, 
                auth_params: Optional[dict] = None, 
                out: Optional[list] = None):
    

    http = urllib3.PoolManager()
    chunk_size = 1024
    download_url = url
    download_file = ""
    signed_access_ids = ['az-blobstore-signed']
    
    
    if is_drs_url(url):
        # parse the drs url to the resource url
        try:
            drs_server, drs_host = get_object_info_url_from_drs(url)
            drs_client = get_drs_client(drs_server)
            object_id = url.split('/')[-1]
            object_info = drs_client.get_object_info(object_id)
        except HTTPError as e:
            if e.response.status_code == 401:
                #if their PAT and email is set, try to log into the appropriate DRS server an retry getting the object
                if email and personal_access_token:
                    try:
                        if drs_host not in oauth_token.keys():
                            oauth_token[drs_host] = login(email,personal_access_token,auth_params=auth_params,drs_url=drs_server)
                        #create a new client with token and get object info with the new client
                        drs_client = get_drs_client(drs_server,oauth_token[drs_host])
                        object_info = drs_client.get_object_info(object_id)
                    except:
                        click.secho("Access Denied", fg='red')
                else:
                    click.secho(f"Authorization required for {drs_server}. Please login or configure your email and personal access token", fg='red')
                    return
            elif e.response.status_code == 404:
                click.secho(f"DRS object with id {object_id} does not exist", fg='red')
                return
            elif e.response.status_code == 403:
                click.secho("Access Denied", fg='red')
                return
            else:
                click.secho("There was an error getting object info from the DRS Client", fg='red')
                return
        except Exception as e:
            return
            

        if "access_methods" in object_info.keys():
            for access_method in object_info["access_methods"][0]:
                if access_method.get('access_id', None):
                    if access_method['access_id'] in signed_access_ids:
                        click.echo("found signed access_id @ access_method level")
                        try:
                            object_access = drs_client.get_object_access(object_id, access_method['access_id'])
                        except HTTPError as e:
                            click.echo(e)
                            return
                        download_url = object_access['url']
                        break

                # if we have an https, use that
                if access_method['type'] == 'https':
                    download_url = access_method['access_url']['url']
                    download_file = download_url.split('/')[-1]
                    break
        else:
            return  # next page token, just return
    else:
        click.secho(f'"{url}" is not a valid DRS url',fg='red')
        return

    try:
        download_stream = http.request('GET', download_url, preload_content=False)
    except:
        click.secho("There was an error downloading " + download_url, fg='red')
        return

    data = handle_file_response(download_file,download_stream.read())
    
    if out is not None: 
        output_lock.acquire()
        out = out.append(file_to_dataframe(download_file,data))
        output_lock.release()

    else:
        with open(f"{output_dir}/{download_file}", 'wb+') as dest:
            dest.write(data.encode('utf-8'))


def download_files(urls, output_dir=downloads_directory, oauth_token: Optional[str] = None, email: Optional[str] = None, personal_access_token: Optional[str] = None, auth_params: Optional[dict] = None, out=None):
    download_threads = []
    
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    for url in urls:
        download = threading.Thread(target=download_file(url, output_dir, 
                                    oauth_token=oauth_token, 
                                    email=email,
                                    personal_access_token=personal_access_token,
                                    auth_params=auth_params,
                                    out=out), name=url)
        download.daemon = True
        download_threads.append(download)
        download.start()

    for thread in download_threads:
        thread.join()

    if out is None:
        click.secho("Download Complete into " + output_dir, fg='green')


import click
import yaml
import json

from covid_cloud.constants import *
from covid_cloud.cli.utils import *


class ConfigHelp(click.Command):
    def format_help(self, ctx, formatter):
        click.echo("Usage: covid-cloud config [OPTIONS] KEY VALUE")
        click.echo("Options:")
        click.echo(" --help  Show this message and exit.")
        click.echo("Accepted Keys: ")
        for key in ACCEPTED_CONFIG_KEYS:
            click.echo(f"\t{key}")


@click.command(cls=ConfigHelp)
@click.pass_context
@click.argument("key")
@click.argument("value", required=False, default=None,nargs=-1)
def config(ctx, key, value):
    if key == 'list' and not value:
        click.echo(json.dumps(ctx.obj, indent=4))
        return
    elif key not in ACCEPTED_CONFIG_KEYS:
        click.secho(f'{key} is not an accepted configuration key.', fg='red')
        return

    if len(value) == 1:
        value = value[0]
    if len(value) == 2:
        value = {value[0]: value[1]}

    set_config(ctx,key,value)

    with open(config_file_path, 'w') as config_file:
        yaml.dump(ctx.obj, config_file)

    click.echo(json.dumps(ctx.obj, indent=4))



import os

auth_params = {
    "wallet_uri": 'https://wallet.staging.dnastack.com',
    "redirect_uri": "https://92a15358-9be7-4edf-8758-ec87ee1f81e3.mock.pstmn.io/" ,
    "client_id": "9110B5F6-B6E4-41D1-A88A-7BD96524FEA3",
    "client_secret": "5DpuE4tXMHi9SShgOCyUmOjyU5RsDYZGTj9yKtPqCo4e58Q8MEILEz0oybNu3TENphHovg8YbxjuHtZk2tFECaPdtRtrkRNIuUw5XCSYhhOzADiCPqZohFsTfpQkoCmS"
}

wallet_uri = 'https://wallet.staging.dnastack.com' #TODO: replace with prod
redirect_uri = "https://92a15358-9be7-4edf-8758-ec87ee1f81e3.mock.pstmn.io/" #TODO: replace this with a proper DNAstack URI
config_file_path = f"{os.path.expanduser('~')}/.covid-cloud/config.yaml"
cli_directory = f"{os.path.expanduser('~')}/.covid-cloud"
downloads_directory = f"{os.path.expanduser('~')}/.covid-cloud/downloads"

ACCEPTED_CONFIG_KEYS = ['search-url', 'drs-url', 'personal_access_token', 'email', 'oauth_token',
                        'wallet-url', 'client-id', 'client-secret', 'client-redirect-uri']

auth_scopes = "openid drs-object:write drs-object:access search:info search:data search:query"


import requests as req
from urllib.parse import parse_qs
from covid_cloud.constants import *
import click

def login(email, personal_access_token, 
            auth_params = auth_params,
            search_url='', drs_url=''):
    try: 
        session = req.Session()
        # login at /login/token
        session.get(f'{auth_params["wallet_uri"]}/login/token',
                    params={
                        'token': personal_access_token,
                        'email': email
                    },
                    allow_redirects=False)
        

        auth_code_res = session.get(f'{auth_params["wallet_uri"]}/oauth/authorize',
                                    params={
                                        'response_type': 'code',
                                        'client_id': auth_params["client_id"],
                                        'redirect_uri': auth_params["redirect_uri"],
                                        'scope': auth_scopes,
                                        'resource': f'{drs_url}'
                                    },
                                    allow_redirects=False)
        
        auth_code = parse_qs(req.utils.urlparse(auth_code_res.headers['Location']).query)['code'][0]

        auth_token_res = session.post(f'{auth_params["wallet_uri"]}/oauth/token',
                                    data={
                                        'grant_type': 'authorization_code',
                                        'code': auth_code,
                                        'scope': auth_scopes,
                                        'resource': f'{drs_url}',
                                        'client_id': auth_params["client_id"],
                                        'client_secret': auth_params["client_secret"]
                                    })

        json_res = auth_token_res.json()
        click.secho("Login successful!",fg='green')

        return json_res['access_token']
    except:
        click.secho("Login failed!",fg='red')
        return None

    



import click
import datetime as dt
from covid_cloud.cli.utils import has_config, get_config
from covid_cloud.cli.auth.commands import cli_login


def get_oauth_token(ctx):

    if not has_config(ctx,"oauth_token"):
        return
    refresh_token(ctx)
    return get_config(ctx,"oauth_token")

def refresh_token(ctx):

    # there is no good way got telling if the server is authenticated yet,
    # so if the user does not have a token set up, assume that is is public and
    # don't do anything
    ctx.invoke(cli_login,refresh=True)


import click
from covid_cloud.cli.auth.utils import get_oauth_token
from covid_cloud.client import *


@click.group()
@click.pass_context
def tables(ctx):
    pass


@tables.command(name="list")
@click.pass_context
def list_tables(ctx):
    click.echo(search_client.get_tables(ctx.obj['search-url'], get_oauth_token(ctx)))


@tables.command()
@click.pass_context
@click.argument('table_name')
def get(ctx, table_name):
    click.echo(search_client.get_table(ctx.obj['search-url'], table_name, get_oauth_token(ctx)))



from covid_cloud.client import *
from covid_cloud import constants
from pandas import DataFrame
from getpass import getpass
import json


class COVIDCloud: 

    def __init__(self, email=None, personal_access_token=None, search_url=None, auth_params=auth_params):
        self.email = email
        self.personal_access_token = personal_access_token
        self.search_url = search_url
        self.auth_params = auth_params
        self.oauth_token = {}

    def login(self, server=None, email=None, personal_access_token=None):
        
        #if the user did not provide a PAT or email, get that from input
        if not server:
            server = input("Enter your DRS Server: ")

        if not (personal_access_token or self.personal_access_token):
            personal_access_token = getpass("Enter your Personal Access Token (PAT): ")
        
        if not (email or self.email):
            email = input("Enter your Email: ")

        drs_host = get_host(server)

        self.oauth_token[drs_host] = login(email=email if email else self.email,
                                 personal_access_token= personal_access_token if personal_access_token else self.personal_access_token,
                                 auth_params=self.auth_params,
                                 search_url=self.search_url,
                                 drs_url=server)

    def query(self, q, download=False, use_json=False, raw=False):
        return json.loads(query(self.search_url, q, download, use_json, raw, self.oauth_token))

    def list_tables(self):
        return json.loads(get_tables(self.search_url, self.oauth_token))

    def get_table(self, table_name):
        return json.loads(get_table(self.search_url, table_name, self.oauth_token))

    def load(self, urls, output_dir=downloads_directory):
        download_content = []
        download_files(urls, output_dir, self.oauth_token, self.email, self.personal_access_token, self.auth_params, download_content)
        return download_content

    def download(self, urls, output_dir=downloads_directory):
        return download_files(urls, output_dir, self.oauth_token, self.email, self.personal_access_token, self.auth_params)



from covid_cloud.cli.utils import assert_config, get_config
from covid_cloud.client import *
from covid_cloud.client.files_client import download_files
from covid_cloud.cli.auth.utils import get_oauth_token


@click.group()
@click.pass_context
def files(ctx):
    pass


@files.command()
@click.pass_context
@click.argument("urls", required=False, nargs=-1)
@click.option('-o', '--output-dir', required=False, default=downloads_directory, show_default=True)
@click.option('-i', '--input-file', required=False, default=None)
@click.option('-q', '--quiet', is_flag=True, required=False)
def download(ctx, urls, output_dir, input_file, quiet):
    download_urls = []

    if len(urls) > 0:
        download_urls = list(urls)
    elif input_file:
        with open(input_file, 'r') as infile:
            download_urls = filter(None, infile.read().split('\n'))  # need to filter out invalid values
    else:
        if not quiet:
            click.echo("Enter one or more URLs. Press q to quit")

        while True:
            try:
                url = click.prompt("", prompt_suffix="", type=str)
                if url[0] == 'q' or len(url) == 0:
                    break
            except click.Abort:
                break

            download_urls.append(url)
    try:
        download_files(download_urls, output_dir, get_config(ctx,'oauth_token'), get_config(ctx,'email'), get_config(ctx,'personal_access_token'), auth_params)
    except:
        return




import click
from covid_cloud.client.auth import login
from covid_cloud.client import *
from covid_cloud.cli.utils import assert_config, get_config, set_config
import covid_cloud.constants
import yaml
import datetime as dt


@click.group()
@click.pass_context
def auth(ctx):
    pass


@auth.command("login")
@click.pass_context
def cli_login(ctx, refresh=False):
    personal_access_token = get_config(ctx, "personal_access_token")
    email = get_config(ctx, "email")
    auth_params = {
        "client_id" : get_config(ctx, "client-id"),
        "client_secret": get_config(ctx, "client-secret"),
        "redirect_uri": get_config(ctx, "client-redirect-uri"),
        "wallet_uri": get_config(ctx, 'wallet-url')
    }
    

    try:
        access_token = login(email, personal_access_token, 
                            auth_params=auth_params, 
                            search_url=get_config(ctx, 'search-url'),
                            drs_url=get_config(ctx, "drs-url"))
    except Exception as e:
        click.secho(f"There was an error generating an access token: {e}", fg='red')
        return

    # convert the expiry from a relative time to an actual timestamp

    set_config(ctx, 'oauth_token', {get_config(ctx, "drs-url"): access_token})

    # if we're refrshing the token, we don't want the user to see that we're refreshing
    if not refresh:
        click.secho('You are now logged in.', fg='green')



from setuptools import find_packages, setup


setup(
    name='covid_cloud',
    packages=find_packages(exclude=['covid_cloud.cli']),
    version='0.0.25',
    description='DNAstack COVID Cloud CLI Library',
    author='Derek, Joseph, Usanthan',
    license='MIT',
    install_requires=['search-python-client>=0.1.9', 'urllib3==1.26.3', 'requests==2.25.1', 'click==7.1.2']
)



import click
from requests.exceptions import SSLError
from search_python_client.search import SearchClient, DrsClient
from typing import Optional
import re




def handle_client_results(results, search_url):
    try:
        yield from results
    except SSLError:
        click.secho(f"There was an error retrieving the SSL certificate from {search_url}", fg='red')
        return
    except:
        click.secho(f"There was an error querying from {search_url}", fg='red')
        return


def get_search_client(search_url, oauth_token: Optional[str] = None):
    # TODO get new token if expired
    if oauth_token:
        search_client = SearchClient(search_url, wallet=oauth_token)
    else:
        search_client = SearchClient(search_url)

    return search_client


def get_drs_client(drs_url, oauth_token: Optional[str] = None):
    # TODO get new token if expired
    if oauth_token:
        drs_client = DrsClient(drs_url, wallet=oauth_token)
    else:
        drs_client = DrsClient(drs_url)

    return drs_client



from covid_cloud.constants import *
from datetime import datetime
from requests.exceptions import SSLError
import click
import json
from requests import HTTPError
from covid_cloud.client.utils import handle_client_results, get_search_client
from typing import Optional


def query(search_url, q, download=False, use_json=False, raw=False, oauth_token: Optional[str] = None):
    search_client = get_search_client(search_url, oauth_token)

    try:
        results = search_client.search_table(q)
    except HTTPError:
        click.echo(HTTPError.response)
        return

    if use_json or not raw:
        output = json.dumps(list(handle_client_results(results, search_url)), indent=4)
    else:
        outputs = []
        for res in handle_client_results(results, search_url):
            data_row = map(str, res.values())
            outputs.append(",".join(data_row))
        output = "\n".join(outputs)

    if download:
        # TODO: be able to specify output file
        download_file = f"{downloads_directory}/query{str(datetime.now())}{'.json' if use_json or not raw else '.csv'}"
        with open(download_file, "w") as fs:
            fs.write(output)
    else:
        return output


def get_tables(search_url, oauth_token: Optional[str] = None):
    search_client = get_search_client(search_url, oauth_token)

    try:
        tables_iterator = search_client.get_table_list()
    except HTTPError:
        click.echo(HTTPError.response)
        return

    return json.dumps(list(handle_client_results(tables_iterator, search_url)), indent=4)


def get_table(search_url, table_name, oauth_token: Optional[str] = None):
    search_client = get_search_client(search_url, oauth_token)

    try:
        table_info = search_client.get_table_info(table_name)
    except SSLError:
        click.secho(f"Unable to retrieve SSL certificate from {search_url}", fg='red')
        return
    except HTTPError as error:
        if '404' in str(error.response):
            click.secho('Invalid table name. ', fg='red')
        else:
            click.echo(error.response)
        return

    # formatting response to remove unnecessary fields
    results = table_info.to_dict()
    results["name"] = table_info["name"]["$id"]
    results["description"] = table_info["description"]["$id"]

    return json.dumps(results, indent=4)
